local MAJOR = "StarLibLCDText-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LCDText = LibStub:NewLibrary(MAJOR, MINOR)
if not LCDText then return end
local core = LibStub("StarLibCore-1.0")
assert(core, MAJOR .. " requires StarLibCore-1.0")
local error = LibStub("StarLibError-1.0")
assert(error, MAJOR .. " requires StarLibError-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LCDText.__index then
	LCDText.__index = LCDText
end

-- @name StarLCDText:New
-- @usage StarLCDText:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor A StarLibCore object
-- @param rows The number of rows
-- @param cols The number of cols
-- @param yres The y resolution for characters
-- @param xres The x resolution for characters
-- @param layers The number of layers
-- @param errorLevel Error verbosity level
-- @return A new StarLCDText object

function LCDText:New(visitor, rows, cols, yres, xres, layers, errorLevel)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.visitor = visitor
	obj.ROWS = rows
	obj.COLS = cols
	obj.LAYERS = layers

	local n = rows * cols
	
	obj.LayoutFB = {}
	obj.TransitionFB = {}
	for i = 0, layers do
		obj.LayoutFB[i] = core:ResizeText("", n)
		obj.TransitionFB[i] = core:ResizeText("", n)
	end
	obj.DisplayFB = core:ResizeText("", n)
	
	return obj	
	
end

-- @name StarLCDText:Del
-- @usage StarLCDText:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LCDText:Del(lcd)
	
	if lcd and type(lcd) == "table" then
		pool[lcd] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
	else
		pool[self] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
	end
end

function LCDText:TextBlit(row, col, height, width)
	local lr, lc
	local dr, dc
	local p1, p2
	local fb = resize("", LROWS * LCOLS)
	
	local r = row
	while r < self.LROWS and r < row + height do
		local c = col
		while c < self.LCOLS and c < col + width do
			local l = 0
			while l < LAYERS do
				if string.sub(LayoutFB[l], r * LCOLS + c, 1) ~= ' ' then
					fb = core:ReplaceText(fb, r * self.LCOLS + c, LayoutFB[l], r * self.LCOLS + c)
				end
				l = l + 1
			end
			c = c + 1
		end
		r = r + 1
	end

	--/* loop over layout rows */
	lr = row
	while lr < LROWS and lr < row + height do
		--/* transform layout to display row */
		dr = lr;
		--/* sanity check */
		if dr >= 0 and dr < self.DROWS then
		
			--/* loop over layout cols */
			lc = col
			while lc < LCOLS and lc < col + width do
				--/* transform layout to display column */
				dc = lc;
				--/* sanity check */
				if dc >= 0 and dc < self.DCOLS then
					--/* find start of difference */
					if string.sub(DisplayFB, dr * DCOLS + dc, 1) ~= string.sub(fb, lr * LCOLS + lc) then
						--/* find end of difference */
						p1 = dc
						p2 = p1
						while lc < LCOLS and lc < col + width do
							--/* transform layout to display column */
							dc = lc;
							--/* sanity check */
							if dc >= 0 and dc < DCOLS then
								if DisplayFB[dr * DCOLS + dc] ~= fb[lr * LCOLS + lc] then
									p2 = dc;
									eq = 0;
								end
							end
							lc  = lc + 1
						end
						--/* send to display */
						local to = string.sub(DisplayFB, dr * DCOLS + 1)
						local from = string.sub(fb, lr * LCOLS + 1)
						local ret = core:Memcopy(to, from, p2 - p1 + 1)
						if (TextRealBlit) then
							TextRealBlit(self, dr, string.sub(DisplayFB, dr * DCOLS + p1), p2 - p1 + 1)
							--TextRealBlit((LCDText *)visitor_->GetLCD(), dr, p1, DisplayFB + dr * DCOLS + p1, p2 - p1 + 1);
						end
					end
				end
				lc = lc + 1
			end
		end
		lr = lr + 1
	end
end
